#include <cassert>
#include <cstdio>
#include <cstdlib>
#include <ctime>

struct ColumnMajorMatrix{
   int m, n;
   double *value;

   explicit ColumnMajorMatrix(int m_=1, int n_=1)
      : m(m_), n(n_), value(0)
   {
      // note: calloc might not return a well-aligned block; for better performance (at the cost of a little memory)
      // you might want to force it to an aligned location in memory
      value=(double*)std::calloc(m*n, sizeof(double));
   }

   ~ColumnMajorMatrix(void)
   {
      std::free(value);
      value=0;
   }
};

void populate_matrix(ColumnMajorMatrix &A);
void gemm_ijk(const ColumnMajorMatrix &A, const ColumnMajorMatrix &B, const ColumnMajorMatrix &C);

int main(int argc, char **argv)
{
   // command-line arguments: matrix size, number of tests to run for timing
   int n=1; // default
   int numtests=0; // default

   if(argc>1){
      std::sscanf(argv[1], "%d", &n);
      if(n<=0){
         std::printf("error: matrix size must be positive\n");
         std::exit(1);
      }
   }
   if(argc>2){
      std::sscanf(argv[2], "%d", &numtests);
      if(numtests<=0){
         std::printf("error: number of tests must be positive\n");
         std::exit(1);
      }
   }

   // initialize data for tests
   ColumnMajorMatrix A(n,n), B(n,n), C(n,n);
   populate_matrix(A);
   populate_matrix(B);

   // test ijk variant =======================================================
   // first warm up cache
   gemm_ijk(A, B, C);
   // now time it
   clock_t start_time=std::clock();
   for(int t=0; t<numtests; ++t){
      gemm_ijk(A, B, C);
   }
   clock_t end_time=std::clock();
   std::printf("ijk  average ticks per test: %lg\n", (end_time-start_time)/(double)numtests);


   //
   // you now need to add the other variants, and a version with a call to BLAS dgemm
   //


   return 0;
}

// put some nonzero values in the matrix
void populate_matrix(ColumnMajorMatrix &A)
{
   for(int j=0; j<A.n; ++j) for(int i=0; i<A.m; ++i){
      A.value[i+j*A.m]=i-j;
   }
}

void gemm_ijk(const ColumnMajorMatrix &A, const ColumnMajorMatrix &B, const ColumnMajorMatrix &C)
{
   // Real code would instead return an error instead of aborting, but for this test it's fine.
   // Make sure once you have debugged your code to compiled with the -DNDEBUG switch to strip out
   // these assertions when measuring performance.
   assert(A.n==B.m);
   assert(C.m==A.m);
   assert(C.n==B.n);
   for(int i=0; i<C.m; ++i) for(int j=0; j<C.n; ++j){
      double d=0;
      for(int k=0; k<A.n; ++k){
         d+=A.value[i+k*A.m]*B.value[k+j*B.m];
      }
      C.value[i+j*C.m]=d;
   }
}

